#include <emmintrin.h>
#include <xmmintrin.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <algorithm>
#include <math.h>

extern "C" __declspec(dllimport) unsigned long __stdcall GetTickCount();

union v4 {
	struct {
		short x, y, z, w;
	};

	struct {
		int u, v;
	};
};

namespace {
#if 0
void movq(v4& a, const v4& b) {
	a = b;
}

void movq(v4& a, const void *b) {
	a = *(const v4 *)b;
}

void movq(void *a, const v4& b) {
	*(v4 *)a = b;
}

void pand(v4& a, const v4& b) {
	a.u &= b.u;
	a.v &= b.v;
}

void por(v4& a, const v4& b) {
	a.u |= b.u;
	a.v |= b.v;
}

void por(v4& a, const void *b) {
	por(a, *(const v4 *)b);
}

void paddw(v4& a, const v4& b) {
	a.x += b.x;
	a.y += b.y;
	a.z += b.z;
	a.w += b.w;
}

void paddw(v4& a, const void *b) {
	paddw(a, *(const v4 *)b);
}

void paddd(v4& a, const v4& b) {
	a.u += b.u;
	a.v += b.v;
}

void paddd(v4& a, const void *b) {
	paddd(a, *(const v4 *)b);
}

void psubw(v4& a, const v4& b) {
	a.x -= b.x;
	a.y -= b.y;
	a.z -= b.z;
	a.w -= b.w;
}

void psubw(v4& a, const void *b) {
	psubw(a,  *(const v4 *)b);
}

void psubd(v4& a, const v4& b) {
	a.u -= b.u;
	a.v -= b.v;
}

void pmaddwd(v4& a, const v4& b) {
	a.u = (int)a.x*b.x + (int)a.y*b.y;
	a.v = (int)a.z*b.z + (int)a.w*b.w;
}

void pmaddwd(v4& a, const char *b) {
	pmaddwd(a, *(const v4 *)b);
}

void pmullw(v4& a, const v4& b) {
	a.x *= b.x;
	a.y *= b.y;
	a.z *= b.z;
	a.w *= b.w;
}

void pmulhw(v4& a, const v4& b) {
	a.x = (a.x*b.x) >> 16;
	a.y = (a.y*b.y) >> 16;
	a.z = (a.z*b.z) >> 16;
	a.w = (a.w*b.w) >> 16;
}

void pmulhw(v4& a, const void *b) {
	pmulhw(a, *(const v4 *)b);
}

void psraw(v4& a, unsigned bits) {
	a.x >>= bits;
	a.y >>= bits;
	a.z >>= bits;
	a.w >>= bits;
}

void psrad(v4& a, unsigned bits) {
	a.u >>= bits;
	a.v >>= bits;
}

void punpckldq(v4& a, const v4& b) {
	a.v = b.u;
}

void packssdw(v4& a, const v4 b) {
	a.x = (unsigned)(a.u + 32768) >= 0x10000 ? (a.u >> 31)^0x7FFF : a.u;
	a.y = (unsigned)(a.v + 32768) >= 0x10000 ? (a.v >> 31)^0x7FFF : a.v;
	a.z = (unsigned)(b.u + 32768) >= 0x10000 ? (b.u >> 31)^0x7FFF : b.u;
	a.w = (unsigned)(b.v + 32768) >= 0x10000 ? (b.v >> 31)^0x7FFF : b.v;
}

void pshufw(v4& a, const v4 b, unsigned char indices) {
	a.x = ((const short *)&b.x)[(indices >> 0) & 3];
	a.y = ((const short *)&b.x)[(indices >> 2) & 3];
	a.z = ((const short *)&b.x)[(indices >> 4) & 3];
	a.w = ((const short *)&b.x)[(indices >> 6) & 3];
}

#else

void movq(v4& a, const v4& b) {
	_mm_storel_epi64((__m128i *)&a, _mm_loadl_epi64((__m128i *)&b));
//	a = b;
}

void movq(v4& a, const void *b) {
	_mm_storel_epi64((__m128i *)&a, _mm_loadl_epi64((__m128i *)b));
//	a = *(const v4 *)b;
}

void movq(void *a, const v4& b) {
	_mm_storel_epi64((__m128i *)a, _mm_loadl_epi64((__m128i *)&b));
//	*(v4 *)a = b;
}

void pand(v4& a, const v4& b) {
//	a.u &= b.u;
//	a.v &= b.v;
	_mm_storel_epi64((__m128i *)&a, _mm_and_si128(_mm_loadl_epi64((__m128i *)&a), _mm_loadl_epi64((__m128i *)&b)));
}

void por(v4& a, const v4& b) {
//	a.u |= b.u;
//	a.v |= b.v;
	_mm_storel_epi64((__m128i *)&a, _mm_or_si128(_mm_loadl_epi64((__m128i *)&a), _mm_loadl_epi64((__m128i *)&b)));
}

void por(v4& a, const void *b) {
	por(a, *(const v4 *)b);
}

void paddw(v4& a, const v4& b) {	
	_mm_storel_epi64((__m128i *)&a, _mm_add_epi16(_mm_loadl_epi64((__m128i *)&a), _mm_loadl_epi64((__m128i *)&b)));
//	a.x += b.x;
//	a.y += b.y;
//	a.z += b.z;
//	a.w += b.w;
}

void paddw(v4& a, const void *b) {
	paddw(a, *(const v4 *)b);
}

void paddd(v4& a, const v4& b) {
	_mm_storel_epi64((__m128i *)&a, _mm_add_epi32(_mm_loadl_epi64((__m128i *)&a), _mm_loadl_epi64((__m128i *)&b)));
//	a.u += b.u;
//	a.v += b.v;
}

void paddd(v4& a, const void *b) {
	paddd(a, *(const v4 *)b);
}

void psubw(v4& a, const v4& b) {
	_mm_storel_epi64((__m128i *)&a, _mm_sub_epi16(_mm_loadl_epi64((__m128i *)&a), _mm_loadl_epi64((__m128i *)&b)));
//	a.x -= b.x;
//	a.y -= b.y;
//	a.z -= b.z;
//	a.w -= b.w;
}

void psubw(v4& a, const void *b) {
	psubw(a,  *(const v4 *)b);
}

void psubd(v4& a, const v4& b) {
	_mm_storel_epi64((__m128i *)&a, _mm_sub_epi32(_mm_loadl_epi64((__m128i *)&a), _mm_loadl_epi64((__m128i *)&b)));
//	a.u -= b.u;
//	a.v -= b.v;
}

void pmaddwd(v4& a, const v4& b) {
	_mm_storel_epi64((__m128i *)&a, _mm_madd_epi16(_mm_loadl_epi64((__m128i *)&a), _mm_loadl_epi64((__m128i *)&b)));
//	a.u = (int)a.x*b.x + (int)a.y*b.y;
//	a.v = (int)a.z*b.z + (int)a.w*b.w;
}

void pmaddwd(v4& a, const char *b) {
	pmaddwd(a, *(const v4 *)b);
}

void pmullw(v4& a, const v4& b) {
	_mm_storel_epi64((__m128i *)&a, _mm_mullo_epi16(_mm_loadl_epi64((__m128i *)&a), _mm_loadl_epi64((__m128i *)&b)));
//	a.x *= b.x;
//	a.y *= b.y;
//	a.z *= b.z;
//	a.w *= b.w;
}

void pmulhw(v4& a, const v4& b) {
	_mm_storel_epi64((__m128i *)&a, _mm_mulhi_epi16(_mm_loadl_epi64((__m128i *)&a), _mm_loadl_epi64((__m128i *)&b)));
//	a.x = (a.x*b.x) >> 16;
//	a.y = (a.y*b.y) >> 16;
//	a.z = (a.z*b.z) >> 16;
//	a.w = (a.w*b.w) >> 16;
}

void pmulhw(v4& a, const void *b) {
	pmulhw(a, *(const v4 *)b);
}

void psraw(v4& a, unsigned bits) {
	_mm_storel_epi64((__m128i *)&a, _mm_srai_epi16(_mm_loadl_epi64((__m128i *)&a), bits));
//	a.x >>= bits;
//	a.y >>= bits;
//	a.z >>= bits;
//	a.w >>= bits;
}

void psrad(v4& a, unsigned bits) {
	_mm_storel_epi64((__m128i *)&a, _mm_srai_epi32(_mm_loadl_epi64((__m128i *)&a), bits));
//	a.u >>= bits;
//	a.v >>= bits;
}

void punpckldq(v4& a, const v4& b) {
	_mm_storel_epi64((__m128i *)&a, _mm_unpacklo_epi32(_mm_loadl_epi64((__m128i *)&a), _mm_loadl_epi64((__m128i *)&b)));
//	a.v = b.u;
}

void packssdw(v4& a, const v4& b0) {
	const v4 b(b0);
//	_mm_storel_epi64((__m128i *)&a, _mm_packs_epi16(_mm_loadl_epi64((__m128i *)&a), _mm_loadl_epi64((__m128i *)&b)));

	a.x = (unsigned)(a.u + 32768) >= 0x10000 ? (a.u >> 31)^0x7FFF : a.u;
	a.y = (unsigned)(a.v + 32768) >= 0x10000 ? (a.v >> 31)^0x7FFF : a.v;
	a.z = (unsigned)(b.u + 32768) >= 0x10000 ? (b.u >> 31)^0x7FFF : b.u;
	a.w = (unsigned)(b.v + 32768) >= 0x10000 ? (b.v >> 31)^0x7FFF : b.v;
}

template<int indices>
void pshufw(v4& a, const v4& b0) {
	const v4 b(b0);

	_mm_storel_epi64((__m128i *)&a, _mm_shufflelo_epi16(_mm_loadl_epi64((__m128i *)&b), indices));
//	a.x = ((const short *)&b.x)[(indices >> 0) & 3];
//	a.y = ((const short *)&b.x)[(indices >> 2) & 3];
//	a.z = ((const short *)&b.x)[(indices >> 4) & 3];
//	a.w = ((const short *)&b.x)[(indices >> 6) & 3];
}

#endif
}

const int BITS_INV_ACC = 4;
const int SHIFT_INV_ROW = 16 - BITS_INV_ACC;
const int SHIFT_INV_COL = 1 + BITS_INV_ACC;
const int RND_INV_ROW = 1024*(6 - BITS_INV_ACC);
const int RND_INV_COL = 16 * (BITS_INV_ACC - 3);
const int RND_INV_CORR = RND_INV_COL - 1;

__declspec(align(16)) const short one_corr[4] = {
	1, 1, 1, 1
};

__declspec(align(16)) const int round_inv_row[2] = {
	RND_INV_ROW, RND_INV_ROW
};

__declspec(align(16)) const short round_inv_col[4] = {
	RND_INV_COL, RND_INV_COL, RND_INV_COL, RND_INV_COL
};

__declspec(align(16)) const short round_inv_corr[4] = {
	RND_INV_CORR, RND_INV_CORR, RND_INV_CORR, RND_INV_CORR
};

__declspec(align(16)) const short tg_1_16[4] = {
	13036, 13036, 13036, 13036
};

__declspec(align(16)) const short tg_2_16[4] = {
	27146, 27146, 27146, 27146
};

__declspec(align(16)) const short tg_3_16[4] = {
	-21746, -21746, -21746, -21746
};

__declspec(align(16)) const short cos_4_16[4] = {
	-19195, -19195, -19195, -19195
};

#if 0
int rounders[8][2] = {
#define X(v) (v), (v)
	X(RND_INV_ROW + 65536 - 2048),
	X(RND_INV_ROW + 3755),
	X(RND_INV_ROW + 2472),
	X(RND_INV_ROW + 1361),
	X(RND_INV_ROW),
	X(RND_INV_ROW - 1139),
	X(RND_INV_ROW - 1024),
	X(RND_INV_ROW - 1301),
};

short tab_i_04[32] = {
	16384,  16384,  16384, -16384,
	21407,   8867,   8867, -21407,
	16384, -16384,  16384,  16384,
	-8867,  21407, -21407,  -8867,
	22725,  12873,  19266, -22725,
	19266,   4520,  -4520, -12873,
	12873,   4520,   4520,  19266,
	-22725,  19266, -12873, -22725,
};

short tab_i_17[32] = {
	22725,  22725,  22725, -22725,
	29692,  12299,  12299, -29692,
	22725, -22725,  22725,  22725,
	-12299,  29692, -29692, -12299,
	31521,  17855,  26722, -31521,
	26722,   6270,  -6270, -17855,
	17855,   6270,   6270,  26722,
	-31521,  26722, -17855, -31521,
};

short tab_i_26[32] = {
	21407,  21407,  21407, -21407,
	27969,  11585,  11585, -27969,
	21407, -21407,  21407,  21407,
	-11585,  27969, -27969, -11585,
	29692,  16819,  25172, -29692,
	25172,   5906,  -5906, -16819,
	16819,   5906,   5906,  25172,
	-29692,  25172, -16819, -29692,
};

short tab_i_35[32] = {
	19266,  19266,  19266, -19266,
	25172,  10426,  10426, -25172,
	19266, -19266,  19266,  19266,
	-10426,  25172, -25172, -10426,
	26722,  15137,  22654, -26722,
	22654,   5315,  -5315, -15137,
	15137,   5315,   5315,  22654,
	-26722,  22654, -15137, -26722,
};
#else
short tab_i_04[32]={
	16384, 16384, 16386, -16384,
	21407, 8867, 8867, -21407,
	16384, -16384, 16382, 16384,
	-8867, 21407, -21407, -8867,
	22725, 12873, 19266, -22725,
	19266, 4520, -4520, -12873,
	12873, 4520, 4521, 19266,
	-22725, 19266, -12873, -22725,
};

short tab_i_17[32]={
	22725, 22725, 22725, -22725,
	29692, 12299, 12299, -29692,
	22725, -22725, 22725, 22725,
	-12299, 29692, -29691, -12299,
	31521, 17855, 26722, -31521,
	26722, 6270, -6270, -17855,
	17855, 6270, 6270, 26722,
	-31521, 26722, -17855, -31521,
};

short tab_i_26[32]={
	21407, 21407, 21407, -21407,
	27969, 11585, 11586, -27969,
	21407, -21407, 21407, 21407,
	-11585, 27969, -27969, -11585,
	29692, 16819, 25172, -29692,
	25172, 5906, -5906, -16819,
	16819, 5906, 5906, 25172,
	-29692, 25172, -16819, -29692,
};

short tab_i_35[32]={
	19266, 19266, 19266, -19266,
	25172, 10426, 10426, -25172,
	19266, -19266, 19266, 19266,
	-10426, 25172, -25172, -10426,
	26722, 15137, 22654, -26722,
	22654, 5315, -5315, -15137,
	15137, 5315, 5315, 22654,
	-26722, 22654, -15137, -26722,
};

int rounders[8][2] = {
	65553, 65553,
	5809, 5809,
	4521, 4521,
	3311, 3311,
	2050, 2050,
	853, 853,
	1054, 1054,
	699, 699,
};
#endif

void dorow(char *out, const char *inp, const char *table, const int *round) {
	v4 mm0, mm1, mm2, mm3, mm4, mm5, mm6, mm7;

	movq(mm0, inp+0);
	pshufw<0xee>(mm5, mm0);
	punpckldq(mm0, mm0);

	movq(mm3, table);
	pmaddwd(mm3, mm0);
	pmaddwd(mm0, table+16);

	movq(mm1, table+32);
	pmaddwd(mm1, mm5);
	pmaddwd(mm5, table+48);

	movq(mm2, inp+8);
	pshufw<0xee>(mm6, mm2);
	punpckldq(mm2, mm2);

	movq(mm4, table+8);
	pmaddwd(mm4, mm2);
	pmaddwd(mm2, table+24);
	paddd(mm3, mm4);
	paddd(mm0, mm2);

	movq(mm7, table+40);
	pmaddwd(mm7, mm6);
	pmaddwd(mm6, table+56);
	paddd(mm1, mm7);
	paddd(mm5, mm6);

	paddd(mm3, round);
	paddd(mm0, round);

	movq(mm4, mm3);
	psubd(mm3, mm1);
	paddd(mm1, mm4);
	psrad(mm3, SHIFT_INV_ROW);
	psrad(mm1, SHIFT_INV_ROW);

	movq(mm4, mm0);
	paddd(mm0, mm5);
	psubd(mm4, mm5);
	psrad(mm0, SHIFT_INV_ROW);
	psrad(mm4, SHIFT_INV_ROW);

	packssdw(mm1, mm0);
	packssdw(mm4, mm3);
	movq(out, mm1);
	pshufw<0xb1>(mm4, mm4);
	movq(out+8, mm4);
}

void docol(char *inp, const char *out) {
	v4 mm0, mm1, mm2, mm3, mm4, mm5, mm6, mm7;

	v4 x0 = *(const v4 *)(inp + 0);
	v4 x1 = *(const v4 *)(inp + 16);
	v4 x2 = *(const v4 *)(inp + 32);
	v4 x3 = *(const v4 *)(inp + 48);
	v4 x4 = *(const v4 *)(inp + 64);
	v4 x5 = *(const v4 *)(inp + 80);
	v4 x6 = *(const v4 *)(inp + 96);
	v4 x7 = *(const v4 *)(inp + 112);
	v4& y0 = *(v4 *)(out + 0);
	v4& y1 = *(v4 *)(out + 16);
	v4& y2 = *(v4 *)(out + 32);
	v4& y3 = *(v4 *)(out + 48);
	v4& y4 = *(v4 *)(out + 64);
	v4& y5 = *(v4 *)(out + 80);
	v4& y6 = *(v4 *)(out + 96);
	v4& y7 = *(v4 *)(out + 112);

	movq	(mm0,tg_1_16);
	movq	(mm2,tg_3_16);
	movq	(mm3,mm0);
	movq	(mm1,x7);
	movq	(mm6,mm2);
	movq	(mm4,x5);
	pmulhw	(mm0,mm1);
	movq	(mm5,x1);
	pmulhw	(mm2,mm4);
	movq	(mm7,x3);
	pmulhw	(mm3,mm5);
	pmulhw	(mm6,mm7);
	paddw	(mm0,mm5);
	paddw	(mm2,mm4);
	psubw	(mm6,mm4);
	paddw	(mm2,mm7);
	psubw	(mm3,mm1);
//	paddw	(mm0,one_corr);
	paddw	(mm6,mm7);
	movq	(mm1,mm0);
	movq	(mm5,mm3);
	paddw	(mm0,mm2);
	psubw	(mm1,mm2);
	paddw	(mm3,mm6);
	psubw	(mm5,mm6);
	movq	(mm2,mm1);
	paddw	(mm1,mm3);
	movq	(mm4,x0);
	psubw	(mm2,mm3);
	movq	(mm3,x4);
	movq	(mm6,mm2);
	pmulhw	(mm2,cos_4_16);
	movq	(mm7,mm1);
	pmulhw	(mm1,cos_4_16);
	paddw	(mm3,mm4);
//	paddw	(mm3,round_inv_corr);
	psubw	(mm4,x4);
	paddw	(mm2,mm6);
//	por	(mm1,one_corr);
	movq	(mm6,x6);
	paddw	(mm1,mm7);
	pmulhw	(mm6,tg_2_16);
	movq	(mm7,one_corr);
psubw(mm7,mm7);
//	paddw	(mm4,round_inv_col);
	psubw	(mm2,mm7);
	paddw	(mm6,x2);
	paddw	(mm7,mm3);
	paddw	(mm7,mm6);
	psubw	(mm3,mm6);
	movq	(mm6,mm7);
	paddw	(mm7,mm0);
	psubw	(mm6,mm0);
	psraw	(mm7,SHIFT_INV_COL);
	movq	(mm0,mm3);
	psraw	(mm6,SHIFT_INV_COL);
	movq	(y0,mm7);
	paddw	(mm3,mm5);
	movq	(mm7,x2);
	psubw	(mm0,mm5);
	movq	(y7,mm6);
	psraw	(mm3,SHIFT_INV_COL);
	pmulhw	(mm7,tg_2_16);
	psraw	(mm0,SHIFT_INV_COL);
	movq	(y3,mm3);
	movq	(mm6,mm4);
//	psubw	(mm6,one_corr);
	psubw	(mm7,x6);
	movq	(mm3,mm1);
	movq	(y4,mm0);
	paddw	(mm4,mm7);
	psubw	(mm6,mm7);
	paddw	(mm3,mm4);
	psubw	(mm4,mm1);
	psraw	(mm3,SHIFT_INV_COL);
	movq	(mm5,mm6);
	psraw	(mm4,SHIFT_INV_COL);
	paddw	(mm6,mm2);
	psubw	(mm5,mm2);
	movq	(y1,mm3);
	psraw	(mm6,SHIFT_INV_COL);
	movq	(y6,mm4);
	psraw	(mm5,SHIFT_INV_COL);
	movq	(y2,mm6);
	movq	(y5,mm5);
}

void idct(short *out, const short *src) {
	dorow((char *)out +   0, (const char *)src +   0, (const char *)tab_i_04, rounders[0]);
	dorow((char *)out +  16, (const char *)src +  16, (const char *)tab_i_17, rounders[1]);
	dorow((char *)out +  32, (const char *)src +  32, (const char *)tab_i_26, rounders[2]);
	dorow((char *)out +  48, (const char *)src +  48, (const char *)tab_i_35, rounders[3]);
	dorow((char *)out +  64, (const char *)src +  64, (const char *)tab_i_04, rounders[4]);
	dorow((char *)out +  80, (const char *)src +  80, (const char *)tab_i_35, rounders[5]);
	dorow((char *)out +  96, (const char *)src +  96, (const char *)tab_i_26, rounders[6]);
	dorow((char *)out + 112, (const char *)src + 112, (const char *)tab_i_17, rounders[7]);
	docol((char *)out +   0, (const char *)out +   0);
	docol((char *)out +   8, (const char *)out +   8);
}

void permute(short *tmp, const short *src) {
	for(int i=0; i<64; i+=8) {
		tmp[i+0] = src[i+0];
		tmp[i+2] = src[i+1];
		tmp[i+4] = src[i+2];
		tmp[i+6] = src[i+3];
		tmp[i+1] = src[i+4];
		tmp[i+3] = src[i+5];
		tmp[i+5] = src[i+6];
		tmp[i+7] = src[i+7];
	}
}

inline short fastround(double v) {
	return v>0 ? (short)(v+0.5) : -(short)(0.5-v);
}

double	basis[8][8];

void initref() {
	// init DCT basis function lookup table
	for(int a=0; a<8; ++a) {
		for(int b=0; b<8; ++b) {
			double tmp = cos(((a+a+1)*b) * (3.14159265358979323846 / 16.0));
			if(b==0)
				tmp /= sqrt(2.0);
			basis[a][b] = tmp * 0.5;
		}
	}
}

void reffdct(short dst[64], const short src[64]) {
	int x,y;
	double tmp;
	double res[64];
		
	// row transform
	for (y=0; y<8; y++) {
		for (x=0; x<8; x++) {
			tmp = (double) src[y*8+0] * basis[0][x]
				+ (double) src[y*8+1] * basis[1][x]
				+ (double) src[y*8+2] * basis[2][x]
				+ (double) src[y*8+3] * basis[3][x]
				+ (double) src[y*8+4] * basis[4][x]
				+ (double) src[y*8+5] * basis[5][x]
				+ (double) src[y*8+6] * basis[6][x]
				+ (double) src[y*8+7] * basis[7][x];
			res[y*8+x] = tmp;
		}
	}
		
	// column transform
	for (x=0; x<8; x++) {
		for (y=0; y<8; y++) {
			tmp = res[0*8+x] * basis[0][y]
				+ res[1*8+x] * basis[1][y]
				+ res[2*8+x] * basis[2][y]
				+ res[3*8+x] * basis[3][y]
				+ res[4*8+x] * basis[4][y]
				+ res[5*8+x] * basis[5][y]
				+ res[6*8+x] * basis[6][y]
				+ res[7*8+x] * basis[7][y];
			dst[y*8+x] = fastround(tmp);
		}
	}	

	for(int i=0; i<64; ++i) {
		int r = dst[i];

		if (r < -2048)
			r = -2048;
		else if (r > 2047)
			r = 2047;

		dst[i] = r;
	}
}

void refidct(short dst[64], const short src[64]) {
	int x,y;
	double tmp;
	double res[64];
		
	// row transform
	for (y=0; y<8; y++) {
		double *t = res + y;

		for (x=0; x<8; x++) {
			tmp = (double) src[0] * basis[x][0]
				+ (double) src[1] * basis[x][1]
				+ (double) src[2] * basis[x][2]
				+ (double) src[3] * basis[x][3]
				+ (double) src[4] * basis[x][4]
				+ (double) src[5] * basis[x][5]
				+ (double) src[6] * basis[x][6]
				+ (double) src[7] * basis[x][7];
			*t = tmp;
			t += 8;
		}

		src += 8;
	}
		
	// column transform
	for (x=0; x<8; x++) {
		const double *u = res + x*8;
		for (y=0; y<8; y++) {
			tmp = u[0] * basis[y][0]
				+ u[1] * basis[y][1]
				+ u[2] * basis[y][2]
				+ u[3] * basis[y][3]
				+ u[4] * basis[y][4]
				+ u[5] * basis[y][5]
				+ u[6] * basis[y][6]
				+ u[7] * basis[y][7];
			dst[y*8+x] = fastround(tmp);
		}
	}

	for(int i=0; i<64; ++i) {
		int r = dst[i];

		if (r < -256)
			r = -256;
		else if (r > 255)
			r = 255;

		dst[i] = r;
	}
}

class IEEE1180Random {
public:
	IEEE1180Random() : seed(1) {}

	long seed;

	long operator()(long L, long H) {
	  static const double z = (double) 0x7fffffff;

	  long i,j;
	  double x;

	  seed = (seed * 1103515245) + 12345;
	  i = seed & 0x7ffffffe;
	  x = ((double) i) / z;
	  x *= (L+H+1);
	  j = (long)x;
	  return j-L;
	}
};

__declspec(align(16)) short testpats[6][10000][64];
__declspec(align(16))short ref[6][10000][64];

void inittestpats() {
	for(int r=0; r<6; ++r) {
		IEEE1180Random randgen;

		int minpix;
		int maxpix;
		int sign;
		switch(r) {
			case 0: minpix = -256; maxpix = 255; sign = 1; break;
			case 1: minpix = -5; maxpix = 5; sign = 1; break;
			case 2: minpix = -300; maxpix = 300; sign = 1; break;
			case 3: minpix = -256; maxpix = 255; sign = -1; break;
			case 4: minpix = -5; maxpix = 5; sign = -1; break;
			case 5: minpix = -300; maxpix = 300; sign = -1; break;
		}

		for(int i=0; i<10000; ++i) {
			short block[64];

			for(int j = 0; j < 64; ++j)
				block[j] = randgen(-minpix, maxpix) * sign;

			short t[64];
			reffdct(t, block);

			permute(testpats[r][i], t);
			refidct(ref[r][i], t);
		}
	}
}

int trial(int champ) {
__declspec(align(16))	short out[64];

fputc('.', stderr);fflush(stderr);

int worsterr = 0;
int data[6][4];

for(int k=0; k<6; ++k) {
	int maxerr = 0;
	int sumerr[64]={0};
	int sumsqerr[64]={0};

	for(int j=0; j<10000; ++j) {
		idct(out, testpats[k][j]);

		const short *x = out;
		const short *y = ref[k][j];

		for(int i=0; i<64; ++i) {
			int r = x[i];
			if (r < -256)
				r = -256;
			else if (r > 255)
				r = 255;
			int e = r - y[i];
			int ae = abs(e);

			if (ae > maxerr) maxerr = ae;
			sumerr[i] += e;
			sumsqerr[i] += e*e;
		}

#if 0
		for(int i=0; i<64; i+=8) {
			printf("%+6d %+6d %+6d %+6d %+6d %+6d %+6d %+6d | %+6d %+6d %+6d %+6d %+6d %+6d %+6d %+6d\n",
				out[i+0],
				out[i+1],
				out[i+2],
				out[i+3],
				out[i+4],
				out[i+5],
				out[i+6],
				out[i+7],
				ref[j][i+0],
				ref[j][i+1],
				ref[j][i+2],
				ref[j][i+3],
				ref[j][i+4],
				ref[j][i+5],
				ref[j][i+6],
				ref[j][i+7]);
		}
#endif
	}

	int tsumerr = 0;
	int tsumsqerr = 0;
	int wsumerr = 0;
	int wsumsqerr = 0;

	for(int i=0; i<64; ++i) {
		int se = sumerr[i];
		int sse = sumsqerr[i];

		tsumerr += se;
		tsumsqerr += sse;
		if (abs(se) > wsumerr) wsumerr = abs(se);
		if (sse > wsumsqerr) wsumsqerr = sse;
	}

	data[k][0] = wsumsqerr;
	data[k][1] = tsumsqerr;
	data[k][2] = wsumerr;
	data[k][3] = tsumerr;

#if 0
	printf("maxerr=%d, worstsqerr=%.6f, avgsqerr=%.6f, worsterr=%.6f, avgerr=%.6f\n"
		, maxerr
		, wsumsqerr / 10000.0
		, tsumsqerr / 640000.0
		, wsumerr / 10000.0
		, tsumerr / 640000.0
		);
#endif

	worsterr += wsumsqerr * wsumsqerr;
//	worsterr += wsumsqerr;
//	if (wsumsqerr > worsterr)
//		wsumsqerr = worsterr;
}

if (worsterr < champ) {
	printf("---------------------------------------------------------\n");

	printf("short tab_i_04[32]={\n");
	for(int i=0; i<32; i += 4) {
		printf("\t%d, %d, %d, %d,\n"
			, tab_i_04[i+0]
			, tab_i_04[i+1]
			, tab_i_04[i+2]
			, tab_i_04[i+3]
			);
	}
	printf("};\n\n");

	printf("short tab_i_17[32]={\n");
	for(int i=0; i<32; i += 4) {
		printf("\t%d, %d, %d, %d,\n"
			, tab_i_17[i+0]
			, tab_i_17[i+1]
			, tab_i_17[i+2]
			, tab_i_17[i+3]
			);
	}
	printf("};\n\n");

	printf("short tab_i_26[32]={\n");
	for(int i=0; i<32; i += 4) {
		printf("\t%d, %d, %d, %d,\n"
			, tab_i_26[i+0]
			, tab_i_26[i+1]
			, tab_i_26[i+2]
			, tab_i_26[i+3]
			);
	}
	printf("};\n\n");

	printf("short tab_i_35[32]={\n");
	for(int i=0; i<32; i += 4) {
		printf("\t%d, %d, %d, %d,\n"
			, tab_i_35[i+0]
			, tab_i_35[i+1]
			, tab_i_35[i+2]
			, tab_i_35[i+3]
			);
	}
	printf("};\n\n");

	printf("int rounders[8][2] = {\n");
	for(int i=0; i<8; ++i)
		printf("\t%d, %d,\n", rounders[i][0], rounders[i][1]);
	printf("};\n\n");


	fprintf(stderr, "%d < %d:\n", worsterr, champ);
	for(int k=0; k<6; ++k) {
		printf("worstsqerr=%.6f, avgsqerr=%.6f, worsterr=%.6f, avgerr=%.6f\n"
			, data[k][0] / 10000.0
			, data[k][1] / 640000.0
			, data[k][2] / 10000.0
			, data[k][3] / 640000.0
			);

		fprintf(stderr, "worstsqerr=%.6f, avgsqerr=%.6f, worsterr=%.6f, avgerr=%.6f\n"
			, data[k][0] / 10000.0
			, data[k][1] / 640000.0
			, data[k][2] / 10000.0
			, data[k][3] / 640000.0
			);
	}
	fputc('\n', stderr);
	putchar('\n');

	fflush(stdout);
	fflush(stderr);
}

	return worsterr;
}

struct Tables {
	int score;
	short cotab[4*32];
	int roundtab[8*2];
};

void load(const Tables& tab) {
	memcpy(rounders, tab.roundtab, sizeof rounders);
	memcpy(tab_i_04, tab.cotab+0, sizeof(short)*32);
	memcpy(tab_i_17, tab.cotab+32, sizeof(short)*32);
	memcpy(tab_i_26, tab.cotab+64, sizeof(short)*32);
	memcpy(tab_i_35, tab.cotab+96, sizeof(short)*32);
}

void save(Tables& tab) {
	memcpy(tab.roundtab, rounders, sizeof rounders);
	memcpy(tab.cotab+0, tab_i_04, sizeof(short)*32);
	memcpy(tab.cotab+32, tab_i_17, sizeof(short)*32);
	memcpy(tab.cotab+64, tab_i_26, sizeof(short)*32);
	memcpy(tab.cotab+96, tab_i_35, sizeof(short)*32);
}

struct TableSorter {
	bool operator()(const Tables *x, const Tables *y) const {
		return x->score < y->score;
	}
};

Tables tabs[128];
Tables *tabrank[128];

int main() {
	srand(GetTickCount());
	initref();
	inittestpats();

	save(tabs[0]);
	tabs[0].score = trial(0);

	for(int i=1; i<128; ++i)
		tabs[i] = tabs[0];

	for(int i=0; i<128; ++i)
		tabrank[i] = tabs+i;

	int bestscore = 0xfffffff;
	int count = 0;
	for(;;) {
		if ((rand() & 255) > 128) {
			int x, y;

			do {
#if 0
				x = rand() & 127;
				y = rand() & 127;
#else
				x = (int)sqrt((float)rand());
				y = (int)sqrt((float)rand());
#endif
			} while(x >= 96 || y >= 96 || x == y);

			Tables *src1 = tabrank[x];
			Tables *src2 = tabrank[y];
			Tables *dst = tabrank[96 + (rand() & 31)];

			for(int i=0; i<128; ++i)
				dst->cotab[i] = (src1->cotab[i] + src2->cotab[i] + 1) >> 1;
//				dst->cotab[i] = rand() & 1 ? src1->cotab[i] : src2->cotab[i];

			for(int i=0; i<16; i += 2) {
				dst->roundtab[i] = (src1->roundtab[i] + src2->roundtab[i] + 1) >> 1;
//				dst->roundtab[i] = rand() & 1 ? src1->roundtab[i] : src2->roundtab[i];
				dst->roundtab[i+1] = dst->roundtab[i];
			}

			load(*dst);
			dst->score = trial(bestscore);
		} else {
			int k;
			int l;

			do {
				l = (int)sqrt(sqrt((float)rand()));
			} while(l >= 96);

			Tables *src = tabrank[l];
			Tables *dst = tabrank[96 + (rand() & 31)];
			*dst = *src;

if (rand()&1) {
			k = rand () & 127;
			if (rand() & 1) {
				for(int k=0; k<128; ++k)
					dst->cotab[k] += (rand() & 31) - 16;
			} else {
//				k &= 7;

				for(k=0; k<8; ++k) {
					dst->roundtab[k * 2] += (rand() & 31) - 16;
					dst->roundtab[k * 2 + 1] = dst->roundtab[k * 2];
				}
			}

			load(*dst);
			dst->score = trial(bestscore);
} else {
			if (rand() & 1) {
				for(int i=0; i<1; ++i) {
					k = rand() & 127;

//					int delta = (rand() & 31) - 16;
					int delta = (rand() & 31) - 16;

					dst->cotab[k] += delta;

					load(*dst);
					dst->score = trial(bestscore);

					if (dst->score < src->score)
						break;

					dst->cotab[k] -= delta*2;

					load(*dst);
					dst->score = trial(bestscore);

					if (dst->score < src->score)
						break;

					dst->cotab[k] += delta;
				}
			} else {
				k = (rand() & 7) * 2;
				for(int i=0; i<8; ++i) {
					int delta = (rand() & 31) - 16;

					dst->cotab[k] += delta;
					dst->roundtab[k+1] = dst->roundtab[k];

					load(*dst);
					dst->score = trial(bestscore);

					if (dst->score < src->score)
						break;

					dst->cotab[k] -= delta*2;
					dst->roundtab[k+1] = dst->roundtab[k];

					load(*dst);
					dst->score = trial(bestscore);

					if (dst->score < src->score)
						break;

					dst->cotab[k] += delta;
					dst->roundtab[k+1] = dst->roundtab[k];

					k = (k+2) & 14;
				}
			}
}
		}

		std::sort(tabrank, tabrank+128, TableSorter());

		if (tabrank[0]->score < bestscore)
			bestscore = tabrank[0]->score;

#if 0
		if (!(++count & 7)) {
			printf("\n");
			for(int i=0; i<64; ++i) {
				printf("%3d) %d\n", i, tabrank[i]->score);
			}
		}
#endif
	}
}
